/*****************************************************************************
 * Author:   Valient Gough <vgough@pobox.com>
 *
 *****************************************************************************
 * Copyright (c) 2004, Valient Gough
 *
 * This library is free software; you can distribute it and/or modify it under
 * the terms of the GNU General Public License (GPL), as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GPL in the file COPYING for more
 * details.
 *
 */

#include "encfs.h"

#include <botan/botan.h>
#include <botan/lookup.h>
#include <botan/look_pk.h>

#include "Botan.h"
#include "Range.h"
#include "Interface.h"

#include <rlog/rlog.h>

using namespace rlog;
using namespace rel;
using namespace Botan;

#define _REGISTER( NAME, DESC, IFACE, NUM, KEYLEN, BLOCKLEN ) \
static Ptr<Cipher> CONCAT(B_func_,NUM) ( const Interface &iface, int keyLen ) \
{ \
    return Ptr<Cipher>(new BotanCipher( iface, NAME, keyLen )); \
} \
static bool CONCAT(B_reg_, NUM) =  \
   Cipher::Register(NAME, DESC, IFACE, KEYLEN, BLOCKLEN, CONCAT(B_func_,NUM));
#define REGISTER( NAME, DESC, IFACE, KEYLEN, BLOCKLEN ) \
   _REGISTER( NAME, DESC, IFACE, __LINE__, KEYLEN, BLOCKLEN )

REGISTER( "AES", "16 byte block cipher",
	Interface("botan/aes", 1, 0, 0),
	Range(128, 256, 64), Range(16,4096,16) );
REGISTER( "Blowfish", "8 byte block cipher",
	Interface("botan/blowfish", 1, 0, 0),
	Range(128, 448, 64), Range(8,4096,8) );
REGISTER( "Twofish", "8 byte block cipher",
	Interface("botan/twofish", 1, 0, 0),
	Range(128, 256, 64), Range(8,4096,8) ); 
REGISTER( "TripleDES", "8 byte block cipher",
	Interface("botan/tripledes", 1, 0, 0),
	Range(128, 192, 64), Range(8,4096,8) );


// used only for last block if it is not a full block size.
const int CIPHER_ROUNDS = 2;


BotanCipher::BotanCipher( const rel::Interface &iface, const char *cipherName, 
	int keyLen )
{
    _iface = iface;
    _cipherName = cipherName;
    _macName = "HMAC(SHA-1)";

    try
    {
	blockCipher = get_block_cipher( _cipherName );
       	
	MessageAuthenticationCode *mac = get_mac( _macName );
	_macLength = mac->OUTPUT_LENGTH;
	//delete mac;

    } catch (Botan::Exception &ex)
    {
	rError("Error in BotanCipher constructor: %s", ex.what());
    }

    if(keyLen > 0)
    {
	_keyLength = keyLen / 8; // convert to octets
    } else
    {
	Range allowed( blockCipher->MINIMUM_KEYLENGTH,
		blockCipher->MAXIMUM_KEYLENGTH, 
		blockCipher->KEYLENGTH_MULTIPLE);

	// look for the closest key length to 160 bits..
	_keyLength = allowed.closest( 160/8 );
    }

    rDebug("Using cipher %s with key length %i", cipherName, _keyLength);
}

BotanCipher::~BotanCipher()
{
    delete blockCipher;
}

rel::Interface BotanCipher::interface() const
{
    return _iface;
}

/*
    create a new key from the password.
    Use SHA to distribute entropy from the password into the key.

    This algorithm must remain constant for backward compatibility.
*/
CipherKey BotanCipher::newKey(const char *password, int passwdLength)
{
    try
    {
	HashFunction *hash;
	if(_keyLength <= 160)
	{
	    hash = get_hash("SHA-160");
	} else
	if(_keyLength <= 256)
	{
	    hash = get_hash("SHA-160");
	} else
	if(_keyLength <= 384)
	{
	    hash = get_hash("SHA-384");
	} else
	    hash = get_hash("SHA-512");

	SecureVector<byte> data = hash->process( 
		(const unsigned char*)password, passwdLength );
	if((int)data.size() != _keyLength)
	    data.grow_to( _keyLength );

	Ptr<SymmetricKey> key ( new SymmetricKey( data ) );

	delete hash;

	return key;
    } catch(Botan::Exception &e)
    {
	rError("Botan exception: %s", e.what());
	return CipherKey();
    }
}

bool BotanCipher::compareKey( const CipherKey &A, const CipherKey &B ) const
{
    Ptr<SymmetricKey> key1 = A;
    Ptr<SymmetricKey> key2 = B;

    if(key1->length() == key2->length()
	    && memcmp( (byte*)key1.get(), (byte*)key2.get() , 
		key1->length() ) == 0)
    {
	return true;
    } else
	return false;
}

/*
    Create a random key.

    This algorithm can change at any time without affecting backward
    compatibility.
*/
CipherKey BotanCipher::newRandomKey()
{
    Ptr<SymmetricKey> key ( new SymmetricKey( _keyLength ) );

    return key;
}

CipherKey BotanCipher::readKey(const unsigned char *data, 
	const CipherKey &encodingKey_)
{
    try
    {
	Ptr<SymmetricKey> masterKey = encodingKey_;

	unsigned int checksum = 0;
	for(int i=0; i<4; ++i)
	    checksum = (checksum << 8) | (unsigned int)data[i];

	// copy to a temporary and decode
	SecureVector<byte> tmpBuf( _keyLength );
	tmpBuf.copy( data+4, _keyLength );

	streamDecode( tmpBuf.begin(), _keyLength, checksum, masterKey );

	// check results
	unsigned int checksum2 = MAC_32( tmpBuf.begin(), _keyLength, 
		masterKey);

	if(checksum2 != checksum)
	{
	    rDebug("checksum mismatch: expected %u, got %u", checksum, 
		    checksum2);
	    return CipherKey();
	}

	Ptr<SymmetricKey> key ( new SymmetricKey( tmpBuf ) );
	return key;

    } catch( Botan::Exception &ex)
    {
	rError("Botan exception: %s", ex.what());
	return CipherKey();
    }
}

// write out the key followed by a message authentication code..
void BotanCipher::writeKey(const CipherKey &key_, unsigned char *output, 
	const CipherKey &encodingKey_)
{
    rAssert(output != NULL);

    Ptr<SymmetricKey> key = key_;
    Ptr<SymmetricKey> masterKey = encodingKey_;

    unsigned int checksum = MAC_32( key->begin(), key->length(), masterKey );

    try
    {
	SecureVector<byte> tmpBuf( key->length() + 4 );
	tmpBuf.copy( 4, key->begin(), key->length() );

	streamEncode( tmpBuf.begin() + 4, key->length(), checksum, masterKey );
	
	for(int i=0; i<4; ++i)
	{
	    tmpBuf[3-i] = checksum & 0xff;
	    checksum >>= 8;
	}
    } catch( Botan::Exception &ex)
    {
	rError("Botan exception: %s", ex.what());
    }
}

int BotanCipher::keySize() const
{
    return _keyLength;
}

int BotanCipher::encodedKeySize() const
{
    return _keyLength + 4;
}

int BotanCipher::cipherBlockSize() const
{
    return blockCipher->BLOCK_SIZE;
}

uint64_t BotanCipher::MAC_64( const unsigned char *src, int len,
	const CipherKey &key_) const
{
    Ptr<SymmetricKey> key = key_;

    KDF *kdf = get_kdf("KDF2(SHA-1)");
    SymmetricKey mac_key = kdf->derive_key(20, key->bits_of(), "hmac key");

    MessageAuthenticationCode *mac = get_mac( _macName );
    mac->set_key( mac_key );
    SecureVector<byte> out = mac->process( src, len );

    // reduece data to a 64 bye value..
    unsigned char h[8] = {0};
    for(unsigned int i=0; i< out.size(); ++i)
	h[i%8] ^= *((unsigned char*)out+i);

    uint64_t value = (uint64_t)h[0];
    for(int i=1; i<8; ++i)
	value = (value << 8) | (uint64_t)h[i];

    return value;
}

static void flipBytes(unsigned char *buf, int size)
{
    unsigned char revBuf[64];

    while(size)
    {
	int toFlip = sizeof(revBuf);
	if(toFlip < size)
	    toFlip = size;
	for(int i=0; i<toFlip; ++i)
	    revBuf[i] = buf[toFlip - (i+1)];
	memcpy(buf, revBuf, toFlip);
	size -= toFlip;
	buf += toFlip;
    }
    memset(revBuf, 0, sizeof(revBuf));
}

bool BotanCipher::streamCode(unsigned char *buf, int size, 
	uint64_t iv64, const CipherKey &key_, bool enc, int rounds) const
{
    Ptr<SymmetricKey> key = key_;

    byte iv64data[8] = {0};
    uint64_t tmp = iv64;
    for(int i=0; i<8; ++i)
    {
	iv64data[8-i] = tmp & 0xff;
	tmp >>= 8;
    }

    KDF *kdf = get_kdf("KDF2(SHA-1)");
    InitializationVector iv = kdf->derive_key(8, key->bits_of(), iv64data, 8);

    Cipher_Dir dir = enc ? ENCRYPTION : DECRYPTION;

    Pipe pipe( get_cipher( _cipherName + "/CFB" , key->bits_of(), iv, dir) );

    bool result = true;
    for(int round = rounds; round; --round)
    {
	pipe.start_msg();
	pipe.write( (const byte*)buf, size );
	pipe.end_msg();

	pipe.read( (byte*)buf, size );
	
	if( round > 1 )
	{
	    // reverse, so that we can do a second pass the other direction
	    // through the bytes.  That way we make the front of the block
	    // dependent on the other end of the block.
	    //   buf <-- encBuf 
	    flipBytes( buf, size );
	}
    }

    return result;
}

bool BotanCipher::streamEncode(unsigned char *buf, int len,
	uint64_t iv64, const CipherKey &key) const
{
    return streamCode( buf, len, iv64, key, true, CIPHER_ROUNDS );
}

bool BotanCipher::streamDecode(unsigned char *buf, int len,
	uint64_t iv64, const CipherKey &key) const
{
    return streamCode( buf, len, iv64, key, false, CIPHER_ROUNDS );
}

bool BotanCipher::blockCode(unsigned char *buf, int size,
	uint64_t iv64, const CipherKey &key_, bool enc ) const
{
    Ptr<SymmetricKey> key = key_;

    Cipher_Dir dir = enc ? ENCRYPTION : DECRYPTION;

    byte iv64data[8] = {0};
    uint64_t tmp = iv64;
    for(int i=0; i<8; ++i)
    {
	iv64data[8-i] = tmp & 0xff;
	tmp >>= 8;
    }

    KDF *kdf = get_kdf("KDF2(SHA-1)");
    InitializationVector iv = kdf->derive_key(8, key->bits_of(), iv64data, 8);

    Pipe pipe( get_cipher( _cipherName + "/CTS" , key->bits_of(), iv, dir) );

    pipe.start_msg();
    pipe.write( (const byte*)buf, size );
    pipe.end_msg();

    pipe.read( (byte*)buf, size );

    return true;
}

bool BotanCipher::blockEncode(unsigned char *buf, int size, 
	uint64_t iv64, const CipherKey &_key) const
{
    return blockCode( buf, size, iv64, _key, true );
}

bool BotanCipher::blockDecode(unsigned char *buf, int size, 
	uint64_t iv64, const CipherKey &_key) const
{
    return blockCode( buf, size, iv64, _key, false );
}

